;************************************************************************************
COMMENT %

Modified 02/28/07 to avoid use of esi and ebx in counter 3 and CPUID.
Modified 03/04/07 for counter2 & counter3 debugging.
Modified 03/09/07 for repetitions and floating point times.

This file contains three macro pairs for clocking code snippets plus an initialization data macro.
Usage is similar to Michael's macros in timers.asm and counter2.asm, e.g.
;***********************************************
COUNT  		EQU 1000
SKIP   		EQU 4
REPS   		EQU 100
PRIORITY	EQU IDLE_PRIORITY_CLASS
...
	counter_initialization
....	
    counter#_begin  COUNT, SKIP, REPS, PRIORITY
    	repeat REPS
		[code for clocking]
		endm
    counter#_end
    print	str$(eax)								;integer output
    print   left$(real8$(__fp_result),6)			;floating point output

;***********************************************

COUNT specifies the number of clocked passes and SKIP the number of preliminary, unclocked passes intended to cast caches, 
branch predictors, etc. into a reproducible configuration.  Valid PRIORITY constants are (windows.inc):
	NORMAL_PRIORITY_CLASS                equ 20h
	IDLE_PRIORITY_CLASS                  equ 40h
	HIGH_PRIORITY_CLASS                  equ 80h
	REALTIME_PRIORITY_CLASS              equ 100h	(CtlAltDel NOT!)

The standard code for reading the timestamp counter is:
	xor    eax,eax
	cpuid
	rdtsc

My own experience with P1 and P4 boxes doesn't show any benefit from including the cpuid instruction within a clocked loop 
and it adds 450 P4 clocks to an empty loop. Feel free to modify these macros as you will.  Macro variables are accessible in 
application code and may prove helpful in separating wheat and chaff.  

Each reference and code loop for the three algorithms contains the dummy code
		xchg	ecx,ecx	
		xchg	edx,edx	
that I've found useful for the Mean Clock algorithm on a P4 CPU.  To see if they're useful for your CPU, comment them out look
for any changes.

ALGORITHMS:
1.  Mean Clocks
    a. Read clock - T1
    b. Execute N empty reference loops and read clock - T2
    c..Read clock - T3
    d. Execute N loops with embedded code and read clock - T4
    e. Calculate [(T4-T3)-(T2-T1)]/N

2.  Minimum Clocks (quad-loops)
	a. Repeat N times:
	   1.  Read clock - T1(i)
	   2.  Execute 4 empty reference loops and read clock - T2(i)
	   3.  If T2(i)-T1(i) < Trmin then Trmin=T2(i)-T1(i)
	b. Repeat N times:
	   1.  Read clock - T3(i)
	   2.  Execute 4 loops with embedded code and read clock - T4(i)
	   3.  If T4(i)-T3(i) < Tcmin then Tcmin=T4(i)-T3(i)
	c. Calculate [Tcmin - Trmin]/4

3.  Adjunct Mean Clocks (quad-loops)
	a. Repeat N times
	   1.  Read clock - T1(i)
	   2.  Execute 4 empty reference loops and read clock - T2(i)
	   3.  Execute 4 loops with embedded code and read clock - T3(i)
	   4.  Tadj = sum( T3(i) - 2*T2(i) + T1(i))
	b. Calculate Tadj/(4*N)

%%%
;************************************************************************************
counter_initialization	MACRO
.data
ALIGN 8
	__ctr_code_count	dq 0	;Code loop count
	__ctr_ref_count		dq 0	;Reference loop count
	__int_result		dq 0	;Final count/loop
	__fp_result			dq 0

	__ctr_iterations	dd 0	;Input value
	__ctr_skips			dd 0	;Input value
	__ctr_repetitions	dd 0	;Input value
	__ctr_loop_count	dd 0	;Loop counter
	__ctr_pass_flag		dd 0

	__ctr_mbx			dd 0
	__ctr_mcx			dd 0
	__minus_2			dd -2.0

ENDM	;counter_initialization
;************************************************************************************

;************************************************************************************
counter1_begin MACRO iterations:REQ, skips:REQ, repetitions:REQ, priority
LOCAL	label_1A
LOCAL	label_1B
LOCAL	label_1C

.code
ALIGN 4
	mov		__ctr_iterations, iterations
	mov		__ctr_skips, skips
	mov		__ctr_repetitions,repetitions
	
	IFNB <priority>
		invoke	GetCurrentProcess
		invoke	SetPriorityClass, eax, priority
	ENDIF

;Reference loop
		mov		__ctr_pass_flag, 0
		mov		eax, __ctr_skips
		cmp		eax,0
		jnz		@F
		mov		__ctr_pass_flag, 1				;If skips=0
		mov		eax, __ctr_iterations
@@:		mov		__ctr_loop_count, eax

label_1A:
__label_1A equ <label_1A>
		xor		eax, eax						;Serialize
		push	ebx
		cpuid
		pop	ebx
		rdtsc									;T1 = EDX:EAX
		push	edx								;Save T1 on stack
		push	eax	

		mov		ecx, __ctr_loop_count
@@:		push	ecx								;Empty reference loop
;************************************************************************************
;******** null reference code ********
		xchg	ecx,ecx	
		xchg	edx,edx	
;************************************************************************************
		pop		ecx
		sub		ecx,1
		jnz		@B

		rdtsc									;T2 = EDX:EAX
		pop		ecx
		sub		eax, ecx
		pop		ecx
		sbb		edx, ecx						;EDX:EAX = Tref = T2 - T1
		mov		DWORD PTR __ctr_ref_count, eax
		mov		DWORD PTR __ctr_ref_count + 4, edx

		test	__ctr_pass_flag,1
		jnz		@F
		mov		__ctr_pass_flag,1
		mov		eax,__ctr_iterations
		mov		__ctr_loop_count, eax
		jmp		__label_1A
@@:

;Code loop
		mov		__ctr_pass_flag,0
		mov		eax, __ctr_skips
		cmp		eax,0
		jnz		@F		
		mov		__ctr_pass_flag, 1				;If skips=0
		mov		eax, __ctr_iterations
@@:		mov		__ctr_loop_count, eax

	
label_1B:
__label_1B equ <label_1B>
		push	ebx
		xor	eax, eax
		cpuid									;Serialize
		pop	ebx
		rdtsc									;EDX:EAX = T3
		push	edx
		push	eax								;Save T3 on stack
		mov		ecx,__ctr_loop_count
label_1C:										;Start test loop
__label_1C equ	<label_1C>
		push	ecx
		xchg	ecx,ecx	
		xchg	edx,edx	
	
ENDM	;counter1_begin
;************************************************************************************
;******** code inserts here ********
;************************************************************************************
counter1_end MACRO
		pop		ecx
		sub		ecx,1
		jnz		__label_1C

		rdtsc									;EDX:EAX = T4
		pop		ecx
		sub		eax, ecx
		pop		ecx
		sbb		edx, ecx						;EDX:EAX = T4 - T3
		mov		DWORD PTR __ctr_code_count, eax
		mov		DWORD PTR __ctr_code_count + 4, edx

		test	__ctr_pass_flag,1
		jnz		@F
		mov		__ctr_pass_flag,1
		mov		eax,__ctr_iterations
		mov		__ctr_loop_count, eax
		jmp		__label_1B

@@:		invoke	GetCurrentProcess
		invoke	SetPriorityClass, eax, NORMAL_PRIORITY_CLASS

		finit
		fild    __ctr_code_count				;(T4-T3)
		fild	__ctr_ref_count				;(T2-T1)	(T4-T3)
		fsubp	st(1),st						;(T4-T3)-(T2-T1)						
		fild    __ctr_iterations				;N	(T4-T3-T2+T1)
		fild	__ctr_repetitions
		fmulp	st(1),st
		fdivp   st(1),st						;(T4-T3-T2+T1)/N
		fst		__fp_result
		fistp   __int_result

		mov		eax, DWORD PTR __int_result
ENDM	;counter1_end
;************************************************************************************
;************************************************************************************

counter2_begin MACRO iterations:REQ, skips:REQ, repetitions:REQ, priority

LOCAL label_2A
LOCAL label_2B
LOCAL label_2C

.code
ALIGN 4
	mov		__ctr_iterations, iterations
	mov		__ctr_skips, skips
	mov		__ctr_repetitions,repetitions

	mov	DWORD PTR __ctr_code_count, -1			;max for uint comparison
	mov	DWORD PTR __ctr_code_count + 4, -1
	mov	DWORD PTR __ctr_ref_count, -1
	mov	DWORD PTR __ctr_ref_count + 4, -1

	IFNB <priority>
		invoke	GetCurrentProcess
		invoke	SetPriorityClass, eax, priority
	ENDIF

;Reference loop
	mov	__ctr_pass_flag, 0
	mov	eax, __ctr_skips
	cmp	eax,0
	jnz	@F		
	mov	__ctr_pass_flag, 1					;If skips=0
	mov	eax, __ctr_iterations
@@:	mov	__ctr_loop_count, eax

ALIGN 4		
label_2A:									;Reference loop
__label_2A equ <label_2A>
	push	ebx
	xor	eax, eax
	cpuid									;Serialize
	pop	ebx
	rdtsc									;EDX:EAX = T1
	push	edx
	push	eax								;Save T1 on stack

	mov		ecx,4							;quad loop for 1 clock precision
@@:	push	ecx
;************************************************************************************
;******** null reference code ********
;	xchg	ecx,ecx	
;	xchg	edx,edx	
;************************************************************************************
	pop	ecx
	sub	ecx,1
	jnz	@B

	rdtsc									;EDX:EAX = T2
	pop	ecx	
	sub	eax, ecx
	pop	ecx	
	sbb	edx, ecx							;EDX:EAX = T2 - T1

	cmp	edx, DWORD PTR __ctr_ref_count + 4
	jne	@F
	cmp	eax, DWORD PTR __ctr_ref_count
@@:	jnb	@F
	mov	DWORD PTR __ctr_ref_count, eax
	mov	DWORD PTR __ctr_ref_count + 4, edx
@@:	sub	__ctr_loop_count, 1				
	jnz	__label_2A
		
	test	__ctr_pass_flag,1
	jnz	@F
	mov	__ctr_pass_flag,1
	mov	eax,__ctr_iterations
	mov	__ctr_loop_count, eax
	mov	DWORD PTR __ctr_ref_count, -1
	mov	DWORD PTR __ctr_ref_count + 4, -1
	jmp	__label_2A
@@:

;***** Code cycle counting	
	mov	__ctr_pass_flag, 0
	mov	eax, __ctr_skips
	cmp	eax,0
	jnz	@F		
	mov	__ctr_pass_flag, 1				;If skips=0
	mov	eax, __ctr_iterations
@@:	mov	__ctr_loop_count, eax

label_2B:							;Code loop
__label_2B equ <label_2B>
	push	ebx
	xor	eax, eax
	cpuid							;Serialize
	pop	ebx
	rdtsc							;EAX:EDX = T3
	push	edx	
	push	eax						;Save T3 on stack

	mov	ecx,4						;quad loop for 1 clock precision
label_2C:
	__label_2C equ <label_2C>
	push 	ecx
;	xchg	ecx,ecx	
;	xchg	edx,edx	
	
ENDM	;counter2_begin
;************************************************************************************
;******** code inserts here ********
;************************************************************************************
counter2_end MACRO
	pop	ecx
	sub	ecx,1
	jnz	__label_2C

	rdtsc							;EDX:EAX = T4
	pop	ecx
	sub	eax, ecx
	pop	ecx
	sbb	edx, ecx					;EDX:EAX = T4 - T3

	cmp	edx, DWORD PTR __ctr_code_count + 4
	jne	@F
	cmp	eax, DWORD PTR __ctr_code_count
@@:	jnb	@F
	mov	DWORD PTR __ctr_code_count,	eax
	mov	DWORD PTR __ctr_code_count + 4, edx
		
@@:	sub	__ctr_loop_count, 1
	jnz	__label_2B

	test	__ctr_pass_flag,1
	jnz	@F
	mov	__ctr_pass_flag,1
	mov	eax,__ctr_iterations
	mov	__ctr_loop_count, eax
	mov	DWORD PTR __ctr_code_count, -1
	mov	DWORD PTR __ctr_code_count + 4, -1
	jmp	__label_2B
		
@@:
	invoke GetCurrentProcess
	invoke SetPriorityClass, eax, NORMAL_PRIORITY_CLASS

	finit
	fld	__minus_2						;Scaling for /4
	fild	__ctr_code_count			;C		-2
	fild	__ctr_ref_count		;R		 C	-2
	fsubp	st(1),st				;(C-R)		-2
	fscale							;(C-R)/4	-2
	fild	__ctr_repetitions
	fdivp	st(1),st
	fst		__fp_result
	fistp	__int_result
	fcomp	st(0)
	mov	eax, DWORD PTR __int_result
		
ENDM	;counter2_end
;************************************************************************************
;************************************************************************************
counter3_begin MACRO iterations:REQ, skips:REQ, repetitions:REQ, priority

LOCAL	label_3A
LOCAL	label_3B

.code
ALIGN 4
		mov		__ctr_iterations, iterations
		mov		__ctr_skips, skips
		mov		__ctr_repetitions,repetitions

	IFNB <priority>
		invoke	GetCurrentProcess
		invoke	SetPriorityClass, eax, priority
	ENDIF

		xor		eax,eax
		mov		dword ptr __ctr_code_count, eax
		mov		dword ptr __ctr_code_count+4, eax	;Tadj=0

		mov		__ctr_pass_flag, 0				;Loop initializations
		mov		ecx, __ctr_skips
		cmp		ecx,0
		jnz		@F
		mov		__ctr_pass_flag, 1				;If skips=0
		mov		ecx, __ctr_iterations
@@:

label_3A:
__label_3A equ <label_3A>
		xor	eax, eax
		push	ebx
		push	ecx
		cpuid
		pop	ecx
		pop	ebx
		rdtsc									;EDX:EAX = T1
		push	edx
		push	eax								;Save T1 on stack

		push	ecx								;Main loop counter
		mov		ecx,4							;quad loop counter
@@:		push	ecx								;Empty loop
;************************************************************************************
;******** null reference code ********
		xchg	ecx,ecx	
		xchg	edx,edx	
;************************************************************************************
		pop		ecx
		sub		ecx,1
		jnz		@B

		pop		ecx
		rdtsc									;EDX:EAX = T2
		push	edx
		push	eax								;Save T2 on stack

		push	ecx								;Main loop counter
		mov		ecx,4							;quad loop counter

label_3B:
__label_3B equ <label_3B>
		push	ecx
		xchg	ecx,ecx	
		xchg	edx,edx	
ENDM	;counter3_begin
;************************************************************************************
;******** code inserts here ********
;************************************************************************************
counter3_end MACRO
		pop		ecx
		sub		ecx,1
		jnz		__label_3B

		pop		ecx
		rdtsc									;EDX:EAX = T3

		pop		__ctr_mcx
		pop		__ctr_mbx						;MBX:MCX = T2
		sub		eax,__ctr_mcx
		sbb		edx,__ctr_mbx
		sub		eax,__ctr_mcx
		sbb		edx,__ctr_mbx						;EDX:EAX = T3 - 2*T2

		pop		__ctr_mcx
		pop		__ctr_mbx						;MBX:MCX = T1
		add		eax,__ctr_mcx
		adc		edx,__ctr_mbx						;EDX:EAX =T3 - 2*T2 + T1

		add		eax, dword ptr __ctr_code_count
		adc		edx, dword ptr __ctr_code_count+4
		mov		dword ptr __ctr_code_count, eax
		mov		dword ptr __ctr_code_count+4, edx	;Tadj = Tadj + (T3 - 2*T2 + T1)

		sub		ecx,1
		jnz		__label_3A

		test	__ctr_pass_flag,1
		jnz		@F
		mov		__ctr_pass_flag,1
		mov		ecx,__ctr_iterations
		xor		eax,eax
		mov		dword ptr __ctr_code_count, eax
		mov		dword ptr __ctr_code_count+4, eax
		jmp		__label_3A

@@:		invoke	GetCurrentProcess
		invoke	SetPriorityClass, eax, NORMAL_PRIORITY_CLASS
		finit
		fild	__ctr_code_count			;C
		fild	__ctr_iterations		;N	C
		fld		st(0)					;N	N	C
		faddp	st(1),st				;2N	C
		fld		st(0)					;2N	2N	C
		faddp	st(1),st				;4N	C
		fild	__ctr_repetitions
		fmulp	st(1),st
		fdivp	st(1),st				;C/4N
		fst		__fp_result
		fistp	__int_result

		mov		eax, DWORD PTR __int_result
ENDM	;counter3_end


